(* Translation from Eff to Delimcc *)

open Eff_semantics;;
open Delimcc_semantics;;
(*
#load "eff_semantics.cmo";;
#load "delimcc_semantics.cmo";;
*)

module Translation(D:Delimcc) = struct

  type 'a repr  = 'a D.repr
  type 'a res   = 'a D.res

  type ('a,'b) eff = ('a,'b) D.free D.prompt

  type ('a,'b) effh = ((unit -> 'a) -> 'b)

  (* values *)
  let int  = D.int
  let add  = D.add
  let unit = D.unit

  let abs = D.abs

  let op: ('a,'b) eff repr -> ('a -> 'b) repr = fun p ->
    D.(abs (fun v -> sh0 p (fun k -> vl @@ act v k)))
 
  let compose: ('b->'c) repr -> ('a->'b) repr -> ('a->'c) repr = fun fbc fab ->
    D.(abs (fun a -> let_ (fab $$ a) (fun b -> fbc $$ b)))

  let handler: ('a,'b) eff repr ->              (* effect instance *)
               ('c repr -> 'w res) ->           (* val handler *)
               ('a repr * ('b -> 'w) repr -> 'w res) -> (* operation handler *)
               ('c,'w) effh repr =
     fun p valh oph ->
       (* The appearance of the universal type comes from the fact that we don't
          carry the effect type in the type of a computation. Therefore,
          the argument and result types of an effect are existentialized
          in res.
        *)
       let h = D.(absrec @@ fun h freer -> 
         with_free freer 
           (fun r   -> let_ (p_univ r) (fun r -> valh r))
             (* Since the handlers are deep, we compose with k with h *)
           (fun v k -> oph (v,compose h k)))
      in 
       D.(abs (fun th -> 
         let_ (pushpr p (let_ (th $$ unit) (fun r -> vl (ret (i_univ r)))))
         (fun freer -> h $$ freer)))

  let vl   = D.vl
  let let_ = D.let_
  let ($$): ('a -> 'b) repr -> 'a repr -> 'b res = D.($$)

  let newp: unit -> ('a,'b) eff res = D.newpr

  let handle: ('c,'w) effh repr -> 'c res -> 'w res =
    fun h e -> h $$ abs (fun (_:unit repr) -> e)
end
;;

let RDelimcc.V (RDelimcc.B 21) = 
  let module M = Ex1(Translation(RDelimcc)) in M.ans
;;

let _ = print_endline "All Done";;
